/* ************************************************************************** */
/*                                       _____           _                    */
/*  libg1m.h                            |_   _|__  _   _| |__   ___ _   _     */
/*  | Project: libg1m                     | |/ _ \| | | | '_ \ / _ \ | | |    */
/*                                        | | (_) | |_| | | | |  __/ |_| |    */
/*  By: thomas <thomas@touhey.fr>         |_|\___/ \__,_|_| |_|\___|\__, |.fr */
/*  Last updated: 2016/10/31 00:14:27                               |___/     */
/*                                                                            */
/* ************************************************************************** */
#ifndef LIBG1M_H
# define LIBG1M_H
# include <libg1m/bcd.h>
# include <libg1m/fontcharacter.h>
# include <libg1m/color.h>
# include <stdio.h>
# include <stdint.h>
# include <time.h>
# ifdef __cplusplus
extern "C" {
# endif

/* ************************************************************************** */
/*  Errors                                                                    */
/* ************************************************************************** */
/* main enumeration */
typedef enum {
	g1m_noerror,

	/* no stream sent -- if lib made the libc calls, check errno */
	g1m_error_nostream,
	/* could not read from stream */
	g1m_error_noread,
	/* could not seek in stream */
	g1m_error_noseek,

	/* magic or control problem */
	g1m_error_magic,
	/* unexpected EOF */
	g1m_error_eof,
	/* memory allocation problem */
	g1m_error_alloc,

	/* operation not supported for this type */
	g1m_error_op,
} g1m_error_t;

/* Message getting macro */
extern const char *g1m_error_strings[];
# define g1m_strerror(N) g1m_error_strings[N]
# define g1m_geterror(N) g1m_error_strings[N]

/* ************************************************************************** */
/*  MCS-Related types                                                         */
/* ************************************************************************** */
/* MCS file type */
# define g1m_mcstype_picture g1m_mcstype_pict
# define g1m_mcstype_capture g1m_mcstype_capt
typedef enum {
	g1m_mcstype_program = 0x01,
	g1m_mcstype_list = 0x02,
	g1m_mcstype_mat = 0x04,
	g1m_mcstype_pict = 0x08,
	g1m_mcstype_capt = 0x10,
	g1m_mcstype_spreadsheet = 0x20
} g1m_mcsfile_type_t;

/* BCD matrix - real and complex */
typedef struct {
	struct bcd real;
	struct bcd imgn;
	int used;
} g1m_mcs_cell_t;

/* mcs file */
# define g1m_has_password(F) (F)->password[0]
typedef struct {
	/* directory name */
	char dirname[9];

	/* file name */
	char name[9];

	/* file type */
	int type;

	/* for programs: the password */
	char password[9];

	/* for spreadsheets, lists and matrixes */
	unsigned int columns, rows;
	g1m_mcs_cell_t **cells;

	/* for pictures and captures */
	unsigned int width, height;
	uint32_t **image; /* 0x0RGB */
	uint32_t **second_image; /* 0x0RGB */
} g1m_mcsfile_t;

/* mcs list */
typedef struct {
	int file_count;
	int _files_size; /* used internally for reallocs */
	g1m_mcsfile_t **files;
} g1m_mcs_t;

/* version */
typedef struct {
	int major;
	int minor;
	int revision;
} g1m_version_t;

/* ************************************************************************** */
/*  General types                                                             */
/* ************************************************************************** */
/* type */
# define g1m_type_pict g1m_type_picture
# define g1m_type_eactivity g1m_type_eact
# define g1m_type_add_in g1m_type_addin
typedef enum {
	g1m_type_addin = 0x01,
	g1m_type_mcs = 0x02,
	g1m_type_eact = 0x04,
	g1m_type_pict = 0x08
} g1m_type_t;

/* handle */
typedef struct {
	/* g1m type */
	g1m_type_t type;
	/* is it a format adapted for fx-CG? */
	int is_cg;

	/* ADD-IN RELATED DATA */
	/* internal name */
	char internal_name[12];
	/* version */
	g1m_version_t version;
	/* creation date */
	time_t creation_date;

	/* MCS RELATED DATA */
	int mcs_count;
	int _mcs_size; /* used internally for reallocs */
	g1m_mcs_t **mcs;

	/* PICTURE RELATED DATA - also used for add-in icons */
	int width, height;
	uint32_t **pixels; /* 0x0RGB */

	/* TODO: e-activities */
} g1m_t;

/* ************************************************************************** */
/*  Main functions                                                            */
/* ************************************************************************** */
/* open handles */
int g1m_open(g1m_t **handle, const char *path);
int g1m_fopen(g1m_t **handle, FILE *stream);

/* open MCS handle (this one is for `libp7`) */
int g1m_parse_mcsfile_content(g1m_mcsfile_t **handle, FILE *stream,
	int raw_type, const char *filename, const char *dirname,
	uint_fast32_t filesize);

/* free handles */
void g1m_free(g1m_t *handle);
void g1m_free_mcsfile(g1m_mcsfile_t *handle);

# ifdef __cplusplus
}
# endif
#endif /* LIBG1M_H */
