/* ************************************************************************** */
/*                                       _____           _                    */
/*  parse/eact.c                        |_   _|__  _   _| |__   ___ _   _     */
/*  | Project: libg1m                     | |/ _ \| | | | '_ \ / _ \ | | |    */
/*                                        | | (_) | |_| | | | |  __/ |_| |    */
/*  By: thomas <thomas@touhey.fr>         |_|\___/ \__,_|_| |_|\___|\__, |.fr */
/*  Last updated: 2016/11/02 14:47:49                               |___/     */
/*                                                                            */
/* ************************************************************************** */
#include <libg1m/internals.h>

/* prototype for content parsing */
static int g1m_parse_eact_cont(g1m_t *handle, uint8_t *buf, size_t bufsize,
	int depth);

/**
 *	g1m_parse_eact_eact:
 *	Parse an EACT in an EACT, from the subheader.
 *
 *	@arg	handle		the handle.
 *	@arg	buf			the buffer.
 *	@arg	bufsize		the buffer size.
 *	@arg	depth		the current depth.
 *	@return				the error code (0 if ok).
 */

static int g1m_parse_eact_eact(g1m_t *handle, uint8_t *buf, size_t bufsize,
	int depth)
{
	int err;

	/* get e-act subheader */
	if (bufsize < sizeof(struct eact_eactheader)) return (g1m_error_eof);
	struct eact_eactheader ehd;
	memcpy(&ehd, buf, sizeof(struct eact_eactheader));

	/* correct endianess */
	ehd.line_count = be32toh(ehd.line_count);

	/* get the line descriptors
	 *	there is actually a "bonus" line descriptor at the end, but the
	 *	fact that we're using offsets lets us unaware of it. */
	if (bufsize < sizeof(struct eact_eactheader)
		+ (ehd.line_count + 1) * sizeof(struct line_descriptor))
		return (g1m_error_eof);
	struct line_descriptor *lds = (void*)&buf[sizeof(struct eact_eactheader)];

	/* browse the lines */
	log_info("[>%d] %d lines to browse", depth, ehd.line_count);
	if (ehd.line_count)
		lds[0].entry_offset = be32toh(lds[0].entry_offset << 8);
	for (uint_fast32_t i = 0; i < ehd.line_count; i++) {
		/* get line size
		 *	for this, we'll calculate the distance to the next line offset.
		 *	if there is no next line (last line), then the data left will be
		 *	attributed to the line. */
		size_t linesize;
		if (i == ehd.line_count - 1)
			linesize = bufsize - lds[i].entry_offset;
		else {
			/* correct endianess of the next element */
			lds[i + 1].entry_offset = be32toh(lds[i + 1].entry_offset << 8);
			/* then calculate */
			linesize = lds[i + 1].entry_offset - lds[i].entry_offset;
		}

		/* store current info */
		uint_fast8_t entry_type = lds[i].entry_type;
		uint_fast32_t entry_offset = lds[i].entry_offset;

		/* check if buffer is big enough */
		if (entry_offset + linesize > bufsize)
			return (g1m_error_eof);

		/* log data */
		log_info("[>%d][%ld] Is '%s' (0x%02x)", depth, i + 1,
			g1m_get_eact_ltype_string(entry_type), entry_type);
#if LOGLEVEL <= ll_info
		if (entry_type != eact_ltype_content) {
			log_info("[>%d][%ld] Buffer (0x%zxo) is:", depth, i + 1, linesize);
			logm_info(&buf[entry_offset], linesize);
		}
#endif

		/* go recursive */
		if (lds[i].entry_type == eact_ltype_content
		 && (err = g1m_parse_eact_cont(handle, &buf[entry_offset], linesize,
		 	 depth + 1)))
			return (err);
	}

	/* no error */
	return (0);
}

/**
 *	g1m_parse_eact_cont:
 *	Parse an E-Activity content.
 *
 *	@arg	handle		the handle.
 *	@arg	buf			the buffer.
 *	@arg	bufsize		the buffer length.
 *	@arg	depth		the current depth.
 *	@return				the error code (0 if ok).
 */

static int g1m_parse_eact_cont(g1m_t *handle, uint8_t *buf, size_t bufsize,
	int depth)
{
	(void)depth;

	/* read content header */
	size_t could_read = min(bufsize, sizeof(struct eact_contentheader));
	if (could_read < 9) return (g1m_error_eof);
	struct eact_contentheader hd = {};
	memcpy(&hd, buf, could_read);

	/* log info */
	log_info("[>%d] Type is '%.8s'", depth, hd.type);
	log_info("[>%d] Name is '%.16s'", depth, hd.name);

	/* prepare for next */
	if (bufsize == could_read)
		return (0);
	buf += could_read;
	bufsize -= could_read;

	/* if type eact, parse eact */
	if (!strcmp((char*)hd.type, "@EACT"))
		return (g1m_parse_eact_eact(handle, buf, bufsize, depth));

	/* otherwise, no error. */
	return (0);
}

/**
 *	g1m_parse_eact:
 *	Parse an EACT.
 *
 *	Thanks to Julese50/CoiledSpring for his help on e-acts parsing.
 *
 *	@arg	handle		the handle.
 *	@arg	stream		the stream to parse from.
 *	@return				the error code (0 if ok).
 */

int g1m_parse_eact(g1m_t * handle, FILE *stream)
{
	/* parse the header */
	DREAD(hd, eact_header)

	/* correct endianess */
	hd.filesize = be32toh(hd.filesize);
	hd.setup_area_size = be32toh(hd.setup_area_size);
	hd.eact_version = be32toh(hd.eact_version);
	hd.os_version = be32toh(hd.os_version);

	/* find out the size of the setup area */
	log_info("E-Activity version is '%s'.",
		hd.eact_version == EACT_G1E ? "g1e" :
		hd.eact_version == EACT_G2E ? "g2e" : "g3e");
	log_info("Setup area size is 0x%x bytes long.",
		hd.setup_area_size);

	/* skip the setup area */
	SKIP(hd.setup_area_size)

	/* get content buffer */
	size_t bufsize = hd.filesize - sizeof(struct standard_header)
		- sizeof(struct eact_header) - hd.setup_area_size;
	uint8_t buf[bufsize];
	READ(&buf, bufsize)

	/* parse content */
	return (g1m_parse_eact_cont(handle, buf, bufsize, 0));
}
